Index: sys/dev/age/if_age.c
===================================================================
--- sys/dev/age/if_age.c	(revision 245870)
+++ sys/dev/age/if_age.c	(working copy)
@@ -2289,9 +2289,7 @@ age_rxeof(struct age_softc *sc, struct rx_rdesc *r
 	nsegs = AGE_RX_NSEGS(index);
 
 	sc->age_cdata.age_rxlen = AGE_RX_BYTES(le32toh(rxrd->len));
-	if ((status & AGE_RRD_ERROR) != 0 &&
-	    (status & (AGE_RRD_CRC | AGE_RRD_CODE | AGE_RRD_DRIBBLE |
-	    AGE_RRD_RUNT | AGE_RRD_OFLOW | AGE_RRD_TRUNC)) != 0) {
+	if ((status & (AGE_RRD_ERROR | AGE_RRD_LENGTH_NOK)) != 0) {
 		/*
 		 * We want to pass the following frames to upper
 		 * layer regardless of error status of Rx return
@@ -2301,9 +2299,18 @@ age_rxeof(struct age_softc *sc, struct rx_rdesc *r
 		 *  o frame length and protocol specific length
 		 *     does not match.
 		 */
-		sc->age_cdata.age_rx_cons += nsegs;
-		sc->age_cdata.age_rx_cons %= AGE_RX_RING_CNT;
-		return;
+		status |= AGE_RRD_IPCSUM_NOK | AGE_RRD_TCP_UDPCSUM_NOK;
+		if ((status & (AGE_RRD_CRC | AGE_RRD_CODE | AGE_RRD_DRIBBLE |
+		    AGE_RRD_RUNT | AGE_RRD_OFLOW | AGE_RRD_TRUNC)) != 0) {
+			for (count = 0; count < nsegs; count++) {
+				rxd = &sc->age_cdata.age_rxdesc[rx_cons];
+				desc = rxd->rx_desc;
+				desc->len = htole32(((MCLBYTES - ETHER_ALIGN) &
+				    AGE_RD_LEN_MASK) << AGE_RD_LEN_SHIFT);
+				AGE_DESC_INC(rx_cons, AGE_RX_RING_CNT);
+			}
+			return;
+		}
 	}
 
 	pktlen = 0;
@@ -2316,10 +2323,10 @@ age_rxeof(struct age_softc *sc, struct rx_rdesc *r
 		if (age_newbuf(sc, rxd) != 0) {
 			ifp->if_iqdrops++;
 			/* Reuse Rx buffers. */
-			if (sc->age_cdata.age_rxhead != NULL) {
+			desc->len = htole32(((MCLBYTES - ETHER_ALIGN) &
+			    AGE_RD_LEN_MASK) << AGE_RD_LEN_SHIFT);
+			if (sc->age_cdata.age_rxhead != NULL)
 				m_freem(sc->age_cdata.age_rxhead);
-				AGE_RXCHAIN_RESET(sc);
-			}
 			break;
 		}
 
@@ -2383,9 +2390,11 @@ age_rxeof(struct age_softc *sc, struct rx_rdesc *r
 			 */
 			if ((ifp->if_capenable & IFCAP_RXCSUM) != 0 &&
 			    (status & AGE_RRD_IPV4) != 0) {
-				m->m_pkthdr.csum_flags |= CSUM_IP_CHECKED;
-				if ((status & AGE_RRD_IPCSUM_NOK) == 0)
+				if ((status & AGE_RRD_IPCSUM_NOK) == 0) {
+					m->m_pkthdr.csum_flags |=
+					    CSUM_IP_CHECKED;
 					m->m_pkthdr.csum_flags |= CSUM_IP_VALID;
+				}
 				if ((status & (AGE_RRD_TCP | AGE_RRD_UDP)) &&
 				    (status & AGE_RRD_TCP_UDPCSUM_NOK) == 0) {
 					m->m_pkthdr.csum_flags |=
@@ -2411,17 +2420,11 @@ age_rxeof(struct age_softc *sc, struct rx_rdesc *r
 			AGE_UNLOCK(sc);
 			(*ifp->if_input)(ifp, m);
 			AGE_LOCK(sc);
-
-			/* Reset mbuf chains. */
-			AGE_RXCHAIN_RESET(sc);
 		}
 	}
 
-	if (count != nsegs) {
-		sc->age_cdata.age_rx_cons += nsegs;
-		sc->age_cdata.age_rx_cons %= AGE_RX_RING_CNT;
-	} else
-		sc->age_cdata.age_rx_cons = rx_cons;
+	/* Reset mbuf chains. */
+	AGE_RXCHAIN_RESET(sc);
 }
 
 static int
@@ -2460,12 +2463,13 @@ age_rxintr(struct age_softc *sc, int rr_prod, int
 		    (MCLBYTES - ETHER_ALIGN)))
 			break;
 
-		prog++;
 		/* Received a frame. */
 		age_rxeof(sc, rxrd);
 		/* Clear return ring. */
 		rxrd->index = 0;
 		AGE_DESC_INC(rr_cons, AGE_RR_RING_CNT);
+		sc->age_cdata.age_rx_cons += nsegs;
+		sc->age_cdata.age_rx_cons %= AGE_RX_RING_CNT;
 	}
 
 	if (prog > 0) {
